
import argparse
import os
import signal
import typing

from types import FrameType

import pyknic.tasks  # noqa: F401  # force tasks loading

from pyknic.tasks.log import LogTask
from pyknic.tasks.config import ConfigTask

from pyknic.lib.log import Logger
from pyknic.lib.verify import verify_value
from pyknic.lib.tasks.proto import TaskProto
from pyknic.lib.tasks.scheduler.scheduler import Scheduler
from pyknic.lib.tasks.scheduler.chain_source import ChainedTasksSource
from pyknic.lib.tasks.threaded_task import ThreadedTask


class App(TaskProto):

    __instance__ = None

    @verify_value(log_level=lambda x: x >= 0)
    def __init__(
        self,
        log_level: int = 0,
        config_file: typing.Optional[str] = None,
        config_dir: typing.Optional[str] = None
    ):
        TaskProto.__init__(self)
        self.__scheduler = Scheduler(task_timeout=60)
        self.__scheduler_thread = ThreadedTask(self.__scheduler, thread_name="pyknic:scheduler")

        self.__main_source = ChainedTasksSource()
        self.__main_source_thread = ThreadedTask(self.__main_source, thread_name="pyknic:chained_tasks_source")

        self.__log_level_int = log_level
        self.__config_file = config_file
        self.__config_dir = config_dir

    def __set_log_level(self) -> None:
        if self.__log_level_int == 0:
            os.environ[LogTask.__env_var_name__] = "ERROR"
        elif self.__log_level_int == 1:
            os.environ[LogTask.__env_var_name__] = "WARN"
        elif self.__log_level_int == 2:
            os.environ[LogTask.__env_var_name__] = "INFO"
        else:
            os.environ[LogTask.__env_var_name__] = "DEBUG"

    def __set_config_vars(self) -> None:
        if self.__config_file is not None:
            os.environ[ConfigTask.__app_file_config_envar__] = self.__config_file

        if self.__config_dir is not None:
            os.environ[ConfigTask.__app_dir_config_envar__] = self.__config_dir

    def start(self) -> None:
        print('Starting the App')

        self.__set_log_level()
        self.__set_config_vars()

        self.__scheduler_thread.start()
        self.__main_source_thread.start()

        self.__scheduler.subscribe(self.__main_source)
        self.__main_source.execute('log_task')
        self.__main_source.execute('config_task')

        config_result = self.__main_source.wait_for(
            self.__main_source.datalog(), 'config_task'
        )
        config = config_result.result  # type: ignore[union-attr]

        apps_enabled = [str(x) for x in config["pyknic"]["apps"]]
        apps_enabled.sort()

        for app_id in apps_enabled:
            Logger.info(f'Starting an app "{app_id}"')
            self.__main_source.execute(app_id)

    def stop(self) -> None:
        self.__main_source_thread.stop()
        self.__main_source_thread.wait()
        self.__main_source_thread.join()

        self.__scheduler_thread.stop()
        self.__scheduler_thread.wait()
        self.__scheduler_thread.join()

    @classmethod
    def terminate_app(cls, sig: int, frame: typing.Optional[FrameType]) -> None:
        print('Stop Request')
        if App.__instance__:
            App.__instance__.stop()
            App.__instance__ = None

    @classmethod
    def parser(cls) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            prog='pyknic',
            description='A fat program that do different magic'
        )
        parser.add_argument('-v', '--verbose', action='count', default=0, help='more flags more logs')
        parser.add_argument('-c', '--config', type=str, help='configuration file')
        parser.add_argument('-C', '--config-dir', type=str, help='configuration directory')

        return parser

    @classmethod
    def main(cls) -> None:
        args = cls.parser().parse_args()

        signal.signal(signal.SIGINT, cls.terminate_app)

        App.__instance__ = ThreadedTask(
            App(log_level=args.verbose, config_file=args.config, config_dir=args.config_dir),
            thread_name='pyknic:starter'
        )
        App.__instance__.start()
        App.__instance__.wait()


if __name__ == "__main__":
    App.main()



# -*- coding: utf-8 -*-

import pytest

from pyknic.path import root_path
from pyknic.lib.gettext import GetTextWrapper


@pytest.fixture
def gettext() -> GetTextWrapper:
    return GetTextWrapper(root_path / 'locales')



# -*- coding: utf-8 -*-
# pyknic/lib/bellboy/secret_backend.py
#
# Copyright (C) 2026 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import abc
import contextlib
import multiprocessing.shared_memory
import sys
import typing

import keyring
import keyring.errors
import pydantic

from pyknic.lib.fastapi.lobby_fingerprint import LobbyFingerprint
from pyknic.lib.fastapi.models.lobby import LobbyFingerprintModel


class SecretTokenModel(pydantic.BaseModel):
    """This model describes a single credential that is required for a single server connection
    """
    server_fingerprint: LobbyFingerprintModel  # server's fingerprint info
    token: str                                 # server's token  # TODO: better make it with a session key!


class PyknicLobbySecrets(pydantic.BaseModel):
    """This model describes a collection of secrets
    """
    secrets: typing.Dict[str, SecretTokenModel]  # key-value, where a key is a URL of a server to which secret is saved


class SecretBackendImplementationProto(metaclass=abc.ABCMeta):
    """Prototype for secret storage that holds data
    """

    @abc.abstractmethod
    def receive_secrets(self) -> typing.Optional[str]:
        """Return from a storage a json-string, that may be parsed as the :class:`PyknicLobbySecrets` object
        """
        raise NotImplementedError('This method is abstract')

    @abc.abstractmethod
    def save_secrets(self, secrets_json_txt: str) -> None:
        """Save to a storage a json-string, that may be parsed as the :class:`PyknicLobbySecrets` object

        :param secrets_json_txt: secrets to save
        """
        raise NotImplementedError('This method is abstract')


class SecretBackend:
    """Basic implementation for a saving/loading login information.

    :note: this implementation is not concurrent safe, so be sure that you don't run tests/multiple login requests
    at the same time

    # TODO: think of a global lock that may be shared by independent processes
    """

    def __init__(self, backend_implementation: typing.Optional[SecretBackendImplementationProto] = None) -> None:
        """Create a new backend

        :param backend_implementation: backend to use. If no backend is specified, then it will be detected with
        the :meth:`.SecretBackend.detect_backend` method
        """
        self.__backend = backend_implementation if backend_implementation else SecretBackend.detect_backend()

    def get_secrets(self) -> PyknicLobbySecrets:
        """Get the secrets stored in backend"""
        try:
            secrets = self.__backend.receive_secrets()
            if secrets:
                return PyknicLobbySecrets.model_validate_json(secrets)
        except pydantic.ValidationError:
            pass

        return PyknicLobbySecrets(secrets=dict())

    def set_secret(self, url: str, fingerprint: LobbyFingerprint, token: str) -> None:
        """Save secrets in backend.
        """
        secrets = self.get_secrets()
        secrets.secrets[url] = SecretTokenModel(
            server_fingerprint=LobbyFingerprintModel(fingerprint=str(fingerprint)),
            token=token
        )
        self.__backend.save_secrets(secrets.model_dump_json())

    def pop_secret(self, url: str) -> SecretTokenModel:
        secrets = self.get_secrets()
        secret_data = secrets.secrets.pop(url)
        self.__backend.save_secrets(secrets.model_dump_json())
        return secret_data

    def purge(self) -> None:
        """Remove every secret from backend.
        """
        self.__backend.save_secrets(PyknicLobbySecrets(secrets=dict()).model_dump_json())

    @staticmethod
    def detect_backend() -> SecretBackendImplementationProto:
        """Try to find usable backend
        """
        try:
            keytool = KeyringSecretBackendImplementation()
            keytool.receive_secrets()
            return keytool
        except keyring.errors.NoKeyringError:
            return SharedMemorySecretBackend()


class KeyringSecretBackendImplementation(SecretBackendImplementationProto):
    """Keyring backend -- https://github.com/jaraco/keyring
    """

    __keytool_service__ = 'pyknic'
    __keytool_login__ = 'secrets'

    def receive_secrets(self) -> typing.Optional[str]:
        """The :meth:`.SecretBackendImplementationProto.receive_secrets` method implementation
        """
        return keyring.get_password(self.__keytool_service__, self.__keytool_login__)

    def save_secrets(self, secrets_json_txt: str) -> None:
        """The :meth:`.SecretBackendImplementationProto.save_secrets` method implementation
        """
        keyring.set_password(self.__keytool_service__, self.__keytool_login__, secrets_json_txt)


class SharedMemorySecretBackend(SecretBackendImplementationProto):
    """Shared memory backend -- look for the files in the /dev/shm directory.
    """

    __shm_filename__ = '/pyknic-secrets'  # the '/' at the beginning is crucial

    def __shared_memory_args(self) -> typing.Dict[str, typing.Any]:
        """ This is a crutch for older python versions defines extra arguments for the
        multiprocessing.shared_memory.SharedMemory constructor.
        """
        if sys.version_info >= (3, 13, 0):
            return {'track': False}
        return {}

    def receive_secrets(self) -> typing.Optional[str]:
        """The :meth:`.SecretBackendImplementationProto.receive_secrets` method implementation
        """
        with contextlib.suppress(FileNotFoundError):
            secret = multiprocessing.shared_memory.SharedMemory(
                self.__shm_filename__, create=False, **self.__shared_memory_args()
            )
            secret_bytes = bytes(secret.buf)  # type: ignore[arg-type]
            result = secret_bytes.decode()
            secret.close()
            return result

        return None

    def save_secrets(self, secrets_json_txt: str) -> None:
        """The :meth:`.SecretBackendImplementationProto.save_secrets` method implementation
        """
        secret_bytes = secrets_json_txt.encode()

        with contextlib.suppress(FileNotFoundError):
            # force file to be removed!
            sh_mem = multiprocessing.shared_memory.SharedMemory(
                self.__shm_filename__, create=False, **self.__shared_memory_args()
            )
            sh_mem.unlink()
            sh_mem.close()

        sh_mem = multiprocessing.shared_memory.SharedMemory(
            self.__shm_filename__,
            size=len(secret_bytes),
            create=True,
            **self.__shared_memory_args()
        )
        sh_mem.buf[:] = secret_bytes  # type: ignore[index]
        sh_mem.close()




import typing


__default_block_size__ = 4096  # usual number of a block size in common FS

IOGenerator: typing.TypeAlias = typing.Generator[bytes, None, None]  # a common generator that produce some data
IOProducer: typing.TypeAlias = typing.Union[IOGenerator, typing.Iterable[bytes]]  # a producer that produces some data
# it looks like IOGenerator but more suitable for function arguments
IOProcessor: typing.TypeAlias = typing.Callable[[IOProducer], IOGenerator]  # a processor that consumes some data
# and produces something else

IOAsyncGenerator: typing.TypeAlias = typing.AsyncGenerator[bytes, None]  # a common async-generator that produce
# some data
IOAsyncProcessor: typing.TypeAlias = typing.Callable[[IOAsyncGenerator], IOAsyncGenerator]  # a processor that
# consumes some data from async generators and produces something else



# -*- coding: utf-8 -*-
# pyknic/lib/io/clients/s3.py
#
# Copyright (C) 2025 the pyknic authors and contributors
# <see AUTHORS file>
#
# This file is part of pyknic.
#
# pyknic is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyknic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyknic.  If not, see <http://www.gnu.org/licenses/>.

import io
import os
import pathlib
import typing

import minio
import minio.datatypes

from pyknic.lib.registry import register_api
from pyknic.lib.uri import URI, URIQuery
from pyknic.lib.io.clients.proto import DirectoryNotEmptyError
from pyknic.lib.io.clients.virtual_dir import VirtualDirectoryClient, path_to_str
from pyknic.lib.io.clients.collection import __default_io_clients_registry__
from pyknic.lib.verify import verify_value

# TODO: register client with registry!


class _S3ClientSyncImplementation:

    @verify_value(uri=lambda x: x.hostname is not None)
    @verify_value(uri=lambda x: x.query is not None)
    def __init__(self, uri: URI, vd_client: VirtualDirectoryClient):

        self.__uri = uri
        self.__vd_client = vd_client
        self.__client: typing.Optional[minio.Minio] = None
        self.__block_size = 4096

        if self.__uri.query is None:
            raise ValueError('Query parameters for S3 Client must be provided')

        self.__uri_query = URIQuery.parse(self.__uri.query)
        self.__bucket_name = self.__uri_query.single_parameter('bucket_name', str)

        if 'block_size' in self.__uri_query:
            block_size = self.__uri_query.single_parameter('block_size', int)
            if block_size < 4096:
                raise ValueError(f'Block size must be greater than {4096} bytes, got {block_size}')
            self.__block_size = block_size

    def connect(self) -> None:

        connection_uri = URI(
            username=self.__uri.username,
            password=self.__uri.password,
            hostname=self.__uri.hostname,
            port=self.__uri.port,
        )

        secure = self.__uri.scheme != 's3'

        self.__client = minio.Minio(
            str(connection_uri),
            secure=secure,
            access_key=self.__uri_query.single_parameter('access_key', str),
            secret_key=self.__uri_query.single_parameter('secret_key', str),
        )

        try:
            available_buckets = self.__client.list_buckets()
            if self.__bucket_name not in (x.name for x in available_buckets):  # mypy issue
                raise ConnectionError(
                    f'There is no such bucket "{self.__bucket_name}", buckets that are available:'
                )
        except minio.error.S3Error as e:
            raise ConnectionError('Connection error') from e

        if self.__uri.path is not None:
            self.change_directory(self.__uri.path)

    def disconnect(self) -> None:
        # TODO: check if there is ".close" method or any others things that will close urllib3 connections
        self.__client = None

    @verify_value(object_path=lambda x: x.is_absolute())
    def __has_inner_objects(self, object_path: pathlib.PosixPath) -> bool:
        try:
            next(self.__list_generator(object_path))
            return True
        except StopIteration:
            pass

        return False

    @verify_value(object_path=lambda x: x.is_absolute())
    def __has_entry(self, object_path: pathlib.PosixPath, directory_entry: bool = True) -> bool:
        assert(self.__client is not None)

        object_path_str = path_to_str(object_path, relative_path=True)
        if not object_path:
            return False

        try:
            if directory_entry:
                object_path_str += '/'
            self.__client.stat_object(self.__bucket_name, object_path_str)
            return True
        except minio.error.S3Error:
            pass

        return False

    def change_directory(self, path: str) -> str:
        if path in ('.', ''):
            return path_to_str(self.__vd_client.session_path())

        posix_path = pathlib.PosixPath(path)
        if not posix_path.is_absolute():
            posix_path = self.__vd_client.session_path() / posix_path

        posix_path = pathlib.PosixPath(path_to_str(posix_path))

        if path_to_str(posix_path) != '/':
            if not self.__has_entry(posix_path, True):
                raise NotADirectoryError(f'There is no such directory: {path}')

        return path_to_str(self.__vd_client.session_path(posix_path))

    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def make_directory(self, directory_name: str) -> None:
        assert(self.__client is not None)

        new_dir_path = self.__vd_client.entry_path(directory_name)

        if self.__has_entry(new_dir_path, False) or self.__has_entry(new_dir_path, True):
            raise FileExistsError(f'Object {path_to_str(new_dir_path)} already exists')

        self.__client.put_object(
            self.__bucket_name, path_to_str(new_dir_path, relative_path=True) + '/', io.BytesIO(b''), 0
        )

    @verify_value(directory_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def remove_directory(self, directory_name: str) -> None:
        assert(self.__client is not None)

        rm_dir_path = self.__vd_client.entry_path(directory_name)

        if not self.__has_entry(rm_dir_path, True):
            raise FileNotFoundError(f'There is no such directory {path_to_str(rm_dir_path)}')

        if self.__has_inner_objects(rm_dir_path):
            raise DirectoryNotEmptyError(f'There is inner object(s) inside a directory {rm_dir_path}')

        self.__client.remove_object(self.__bucket_name, path_to_str(rm_dir_path, relative_path=True) + '/')

    @verify_value(object_path=lambda x: x.is_absolute())
    def __list_generator(self, list_path: pathlib.PosixPath) -> typing.Generator[str, None, None]:
        assert(self.__client is not None)

        request_path_str = path_to_str(list_path, relative_path=True)
        if request_path_str == '.':
            request_path_str = ''

        def map_items(m: minio.datatypes.Object) -> str:
            return m.object_name[len(request_path_str):].rstrip('/').lstrip('/')  # type: ignore[index]  # mypy issue

        path = path_to_str(list_path, relative_path=True)
        if path:
            path += '/'

        all_entries = self.__client.list_objects(self.__bucket_name, prefix=path)

        relative_names_entries = map(map_items, all_entries)
        filtered_entries = filter(lambda y: y != '', relative_names_entries)

        for x in filtered_entries:
            next_item = pathlib.PosixPath(x).parts[-1].rstrip('/')
            if next_item != request_path_str:
                yield next_item.rstrip('/')

    def list_directory(self) -> typing.Tuple[str, ...]:
        return tuple(self.__list_generator(self.__vd_client.session_path()))

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    def upload_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        assert(self.__client is not None)

        local_file_obj.seek(0)
        data_length = local_file_obj.seek(0, os.SEEK_END)
        local_file_obj.seek(0)

        new_file_path = self.__vd_client.entry_path(remote_file_name)

        if self.__has_entry(new_file_path, False) or self.__has_entry(new_file_path, True):
            raise FileExistsError(f'Object {path_to_str(new_file_path)} already exists')

        self.__client.put_object(
            self.__bucket_name,
            path_to_str(new_file_path, relative_path=True),
            local_file_obj,  # type: ignore[arg-type]  # mypy issue
            data_length
        )

    @verify_value(file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def remove_file(self, file_name: str) -> None:
        assert(self.__client is not None)

        rm_file_path = self.__vd_client.entry_path(file_name)

        entry = self.__has_entry(rm_file_path, False)
        if entry is None:
            raise FileNotFoundError(f'There is no such file {path_to_str(rm_file_path)}')

        self.__client.remove_object(self.__bucket_name, path_to_str(rm_file_path, relative_path=True))

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    @verify_value(local_file_obj=lambda x: x.seekable())
    def receive_file(self, remote_file_name: str, local_file_obj: typing.IO[bytes]) -> None:
        assert(self.__client is not None)

        local_file_obj.seek(0)
        local_file_obj.truncate()
        file_path = self.__vd_client.entry_path(remote_file_name)

        if not self.__has_entry(file_path, False):
            raise FileNotFoundError(f'There is no such file {path_to_str(file_path)}')

        http_request = self.__client.get_object(
            self.__bucket_name, path_to_str(file_path, relative_path=True)
        )

        data = http_request.read(self.__block_size)
        while data != b'':
            local_file_obj.write(data)
            data = http_request.read(self.__block_size)

    @verify_value(remote_file_name=lambda x: len(pathlib.PosixPath(x).parts) == 1)
    def file_size(self, remote_file_name: str) -> int:
        assert(self.__client is not None)

        file_path = self.__vd_client.entry_path(remote_file_name)

        if not self.__has_entry(file_path, False):
            raise FileNotFoundError(f'There is no such file {path_to_str(file_path)}')

        result = self.__client.stat_object(self.__bucket_name, path_to_str(file_path, relative_path=True))
        assert(result.size is not None)
        return result.size


@register_api(__default_io_clients_registry__, "s3s")
@register_api(__default_io_clients_registry__, "s3")
class S3Client(VirtualDirectoryClient):

    def __init__(self, uri: URI) -> None:
        """Create a new client

        :param uri: URI with which this client should be created. If uri has the "path" attribute, then this path
        will be used as a start point
        """

        VirtualDirectoryClient.__init__(self, uri)
        self.__i12n = _S3ClientSyncImplementation(uri, self)

        self._wrap_capability(self.__i12n, 'connect')
        self._wrap_capability(self.__i12n, 'disconnect')
        self._wrap_capability(self.__i12n, 'change_directory')
        self._wrap_capability(self.__i12n, 'list_directory')
        self._wrap_capability(self.__i12n, 'make_directory')
        self._wrap_capability(self.__i12n, 'remove_directory')
        self._wrap_capability(self.__i12n, 'upload_file')
        self._wrap_capability(self.__i12n, 'remove_file')
        self._wrap_capability(self.__i12n, 'receive_file')
        self._wrap_capability(self.__i12n, 'file_size')

    def current_directory(self) -> str:
        """The :meth:`.IOClientProto.current_directory` method implementation."""
        return str(self.session_path())
